/*
 * Copyright (C) 2004  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "XMLExtract.h"


#include <fstream>
#include <string.h>
#include <ctype.h>

#include "lib_conversions.h"

#define XMLExtractDebug 0

int nLineNr;

/**
 */
int goUpInPath(std::string& path,
               std::string  up)
{
  std::string ende = path.substr(path.length()-up.length()+1);
    
  // '/' entfernen
  up = up.substr(1);
  
#if XMLExtractDebug == 1  
  std::cout << nLineNr << " Aufwärts: " << up << "\n";
  std::cout << nLineNr << "   Vergleiche " << up << " und " << ende << ".\n";
#endif  
  if (strcmp(up.c_str(), ende.c_str()) == 0)
  {
    // alles in Ordnung, es geht so hoch wie es vorher runter ging
    if (up.length() == path.length())
      path = "";
    else
      path = path.substr(0, path.length()-up.length()-1);
#if XMLExtractDebug == 1  
    std::cout << nLineNr << "   ok. path = " << path << "\n";
#endif
    return(0);
  }
  else
  {
#if XMLExtractDebug == 1  
    std::cout << "  nicht ok!\n";
#endif
    // In den Pfad tue ich die Fehlermeldung.
    path = "Error in xml, line " + itoStr(nLineNr, ' ', 1) + ": when in " + path + " you cant leave element " + up;
    return(-1);
  }  
}

/**
 * Liefert alles...
 */
int XMLExtract(std::string& buf,
               std::string  filename,
               std::string  section,
               bool         fMitTags)
{
  bool          fLesen         = true;
  bool          fTag           = false;
  int           nTagNameFertig = 0;
  bool          fInSection     = false;

  int           data;
  char          c0     = ' ';
  char          c1     = ' ';
  std::ifstream in;
  std::string   tag;
  std::string   element;
  std::string   path   = "";

  
  nLineNr = 1;
  
#if XMLExtractDebug == 1  
  std::cout << section << " in " << filename << "\n";
#endif      
  in.open(filename.c_str());
  
  if (!in)
  {
    buf = "Error opening " + filename;
    return(-1);
  }
  
  while (fLesen)
  {
    data = in.get();
    
    if (data < 0)
      fLesen = false;
    else
    {
      c0 = (char)data;
      
      if (c0 == '\n')
        nLineNr++;
      
      if (c0 == '<' && fTag == false)
      {
        fTag           = true;
        nTagNameFertig = 0;
        tag            = "";
        element        = "<";
      }
      else if (fTag == true)
      {
        if (nTagNameFertig == 1 && (isspace(c0) != 0) || c0 == '>')
          nTagNameFertig = 2; // Name beendet
        
        if (nTagNameFertig == 0 && isspace(c0) == 0)
          nTagNameFertig = 1; // Name hat zumindest angefangen        

        element += c0;
        
        if (c0 == '>')
        {
          if (c1 == '/' || c1 == '?')
          {
            // ein in sich geschlossenes Element
            // oder das einleitende
            if (fInSection && fMitTags)
              buf += element;
          }
          else if (tag[0] == '/')
          {
            // Elementende
            if (goUpInPath(path, tag))
            {
              // ein Fehler in der XML-Struktur?
              buf = path;
              return(-1);
            }
            
            if (fInSection && path.length() < section.length())
              fInSection = false;
            
            if (fInSection && fMitTags)
              buf += element;
          }
          else
          {
            // Elementanfang. 
            if (fInSection && fMitTags)
              buf += element;
            
            if (path.length())
              path += '.';
            path += tag;

#if XMLExtractDebug == 1  
            std::cout << nLineNr << " Abwärts. Neuer Pfad: " << path << "\n";
#endif
            if (path.compare(section)==0)
              fInSection = true;
          }
                      
          fTag = false;
        }
        else if (nTagNameFertig == 1)
          tag += c0;
        
        c1 = c0;
      }
      else if (fInSection == true)
      {
        buf += c0;
      }
    }
  }
    
  in.close();
  
  return(0);
}
