/*
 * Copyright (C) 2004  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "create.h"

#include <unistd.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <iostream>
#include <string>
#include <dirent.h>     // fr die Verzeichnisoperationen
#include <sys/stat.h>   // Abfragen der Dateiattribute
#include <sys/types.h>

#include "lib_conversions.h"
#include "tools.h"


#define SHOWWHATTODOWITHLINKS 0

/**
 *
 */
bool SiVeCoCreate::isIncluded(regex_t*           regex,
                              const char*        name)
{
  if (regex == (regex_t*)0)
    return(true);
  else
  {
    if (regexec(regex, name, 0, 0, 0) == 0)
    {
      excludedFiles.push_back(name);
      return(false);
    }
    else
      return(true);
  }
}

/**
 * 
 */
int SiVeCoCreate::create(std::string        dir,
                         SimpleXMLTransfer* ofile)
{
  DIR*            Verz;                 // der VerzeichnisStrom
  dirent*         VerzEintr;            // hier werden die Daten aus obigem gegeben
  struct stat     Attribs;              // ein Puffer fr die Dateiattribute
  std::string     currentDir = wd;
  
  // Wieviele Pfadabschnitte gibt es? In einem neuen String
  // fr jeden ein '../' erzeugen und noch eins dazu.
  std::string relOut = "../";
  int         ups   = 0;
  {
    char* chrPtrEnd = &dir[dir.length()-1];
    char* chrPtr    = strchr(&dir[0], '/');

    while (chrPtr != (char*)0 && chrPtr <= chrPtrEnd)
    {
      relOut += "../";
      ups++;
      chrPtr = strchr(chrPtr, '/')+1;
    }
    ups *= 3;
  }
  
#if SHOWWHATTODOWITHLINKS == 1
  std::cout << "SHOWWHATTODOWITHLINKS: dir    = " << dir << "\n";
  std::cout << "SHOWWHATTODOWITHLINKS: relOut = " << relOut << "\n";
#endif          
  
  currentDir.append(dir);
    
  Verz = opendir(currentDir.c_str());
  if (Verz != 0)
  {
    while ((VerzEintr = readdir(Verz)) != NULL)
    {
      std::string currentFile = currentDir;
      currentFile += VerzEintr->d_name;
      lstat(currentFile.c_str(), &Attribs);

      // der relative Pfad dieses Elements
      std::string relName = dir + VerzEintr->d_name;
      // Unter diesem Pfad muss die Datei gesichert werden:
      std::string repfile = targetDir + relName;
            
      if (S_ISLNK(Attribs.st_mode) != 0)
      {
        // Ein Link. Man bernimmt es in das Repository.
        // Dabei merkt man sich, ob das Link im Original ein relatives oder
        // ein absolutes war.
        // Zunchst einmal schauen, wohin das Link zeigt:
        std::string        target = getLinkTarget(currentFile);
        SimpleXMLTransfer* xml    = new SimpleXMLTransfer();
        xml->setName("link");
        xml->addAttribute("name", VerzEintr->d_name);
        if (target.length() == 0)
        {
          std::cout << "Error resolving link (1)\n";
          return(-1);
        }
        if (target[0] == '/')
        {
          // Link ist absolut, kann so bleiben
          xml->addAttribute("abs",  "1");
        }
        else
        {
          // Link ist relativ, muss in ein absolutes umgewandelt werden,
          // wenn es nach auerhalb zeigt.
          // Wenn das Link nach innerhalb zeigt, lsst man es so wie es ist.
          // zeigt es nach innerhalb oder auerhalb?
#if SHOWWHATTODOWITHLINKS == 1
          std::cout << "SHOWWHATTODOWITHLINKS: target = " << target << "\n";
#endif          
          if (target.find(relOut) == 0)
          {         
            target = relPathToAbs(currentDir + target);
            if (target.length() == 0)
            {
              std::cout << "Error resolving link (2)\n";
              return(-1);
            }
          }
          xml->addAttribute("abs",  "0");
        }
        ofile->addChild(xml);
        // Link anlegen
        if (cmd == CMD_CREATE)
        {
          if (symlink(target.c_str(), repfile.c_str()) != 0)
          {
            std::cout << "Error when creating symlink\n";
            return(-1);
          }      
        }
      }
      else
      {
        if (S_ISDIR(Attribs.st_mode) != 0)
        {
          if (strcmp("..", VerzEintr->d_name) != 0 && strcmp(".", VerzEintr->d_name) != 0)
          {
            if (isIncluded(exclDir, relName.c_str()))
            {
              // Normales Verzeichnis. Einfach erzeugen und darin weitermachen.
              if (cmd == CMD_CREATE)
              {              
                if (mkdir(repfile.c_str(), S_IRWXU) != 0)
                {
                  std::cout << "Error when creating directory\n";
                  return(-1);
                }
              }
            
              SimpleXMLTransfer* xml2 = new SimpleXMLTransfer();
              xml2->setName("dir");
              xml2->addAttribute("name", VerzEintr->d_name);

              if (create(relName+"/", xml2) != 0)
              {
                std::cout << "Error\n";
                return(-1);
              }
            
              ofile->addChild(xml2);
            }
          }
        }
        else if (isIncluded(exclFile, relName.c_str()))
        {
          // Normale Datei. Vergleichen, evtl. kopieren...
          // Die entsprechende Datei der vorigen Version:
          std::string reffile = refDir + relName;
                    
          // Ist die Referenzdatei vorhanden?
          if (refDir.length() != 0 && existDat(reffile))
          {
            // Vergleichen, evtl. kopieren...
            std::string command = conf->attribute("cmp_cmd") + " \"" + currentFile + "\" \"" + reffile + "\"";
            int         retVal;
            retVal = system(command.c_str());
            retVal = WEXITSTATUS(retVal);
            
            switch (retVal)
            {
             case 0: // Datei ist unverndert, Link setzen
              if (cmd == CMD_CREATE)
              {
                std::string currentFile = getOriginalFileRel(reffile, refDir);
                if (currentFile.length() == 0)
                {              
                  std::cout << "Error when creating symlink\n";
                  return(-1);
                }
                if (symlink(currentFile.c_str(), repfile.c_str()) != 0)
                {
                  std::cout << "Error when creating symlink\n";
                  return(-1);
                }      
              }
              break;
              
             case 1: // Datei wurde verndert, kopieren
              if (cmd == CMD_CREATE)
              {
                command = "cp -p \"" + currentFile + "\" \"" + repfile + "\"";
                system(command.c_str());
              }              
              changedFiles.push_back(relName);
              break;
              
             default:
              std::cout << "Error: unexpected return value from cmp: " << retVal << "\n";
              return(-1);              
              break;
            }
            
          }
          else
          {
            if (cmd == CMD_CREATE)
            {
              // Datei ist neu, also kopieren.
              std::string command = "cp -p \"" + currentFile + "\" \"" + repfile + "\"";
              system(command.c_str());
            }
            newFiles.push_back(relName);
          }          
        }          
      }
    }
    closedir(Verz);
    return(0);
  }
  else
    return(-1);
}

/**
 * Lists files in the reference directory which are not present in the working directory.
 * Only real files and relative symlinks are interesting.
 */
int SiVeCoCreate::fillDeletedFileList(std::string        dir)
{
  DIR*            Verz;                 // der VerzeichnisStrom
  dirent*         VerzEintr;            // hier werden die Daten aus obigem gegeben
  struct stat     Attribs;              // ein Puffer fr die Dateiattribute
  std::string     currentDir = refDir;
  
  currentDir.append(dir);
    
  Verz = opendir(currentDir.c_str());
  if (Verz != 0)
  {
    while ((VerzEintr = readdir(Verz)) != NULL)
    {
      std::string currentFile = currentDir;
      currentFile += VerzEintr->d_name;
      lstat(currentFile.c_str(), &Attribs);

      // der relative Pfad dieses Elements
      std::string relName = dir + VerzEintr->d_name;
            
      if (S_ISLNK(Attribs.st_mode) != 0)
      {
        // Ein Link.
        // Zunchst einmal schauen, wohin das Link zeigt:
        std::string        target = getLinkTarget(currentFile);
        if (target.length() == 0)
        {
          std::cout << "Error resolving link (3)\n";
          return(-1);
        }
        if (target[0] != '/')
        {
          // Link ist relativ         
          if (existDat(wd+relName) == false)
            deletedFiles.push_back(relName);
        }
      }
      else
      {
        if (S_ISDIR(Attribs.st_mode) != 0)
        {
          if (strcmp("..", VerzEintr->d_name) != 0 && strcmp(".", VerzEintr->d_name) != 0)
          {
            if (fillDeletedFileList(relName+"/") != 0)
            {
              std::cout << "Error\n";
              return(-1);
            }
          }
        }
        else
        {
          // Normale Datei.                    
          if (existDat(wd+relName) == false)
            deletedFiles.push_back(relName);
        }          
      }
    }
    closedir(Verz);
    return(0);
  }
  else
    return(-1);
}

/**
 * 
 */
SiVeCoCreate::SiVeCoCreate(SimpleXMLTransfer* ptr_conf,
                           std::string        workingDir,
                           eCmd               i_cmd)
{
  conf = ptr_conf;
  wd   = workingDir;
  cmd  = i_cmd;
}

/**
 * Create new version or make a diff
 */
int SiVeCoCreate::run(int         nNr,
                      std::string type,
                      std::string label)
{
  int                 nCurrent;    
  int                 nDerivedFrom;
  std::string         ofilename;  
  std::string         content;
  std::string         labelToUse;
  SimpleXMLTransfer*  out = (SimpleXMLTransfer*)0;
  SimpleXMLTransfer*  xml;
  SimpleXMLTransfer*  verboseOut;
  std::ofstream       ofile;  

  // Check if repository exists
  if (existDir(getConfDir(conf, "repository")) == false)
  {
    std::cout << "Error: repository " << getConfDir(conf, "repository") << " does not exist.\n";
    return(-1);
  }

  // Check if working directory exists
  if (existDir(wd) == false)
  {
    std::cout << "Error: workingDirectory " << wd << " does not exist.\n";
    return(-1);
  }
    
  // old version
  // Wenn eine Nummer angegeben wurde, wird diese benutzt,
  // ansonsten die aktuelle.
  nCurrent     = getCurrentVersionNumberInt(conf);
  nDerivedFrom = isDerivedFrom(conf, wd);
  
  if (nNr != -1                 // Wenn als Argument eine Nummer bergeben wurde
      && cmd == CMD_CREATE      // und man eine neue Version erstellen soll
      && nDerivedFrom != -1     // und bekannt ist, wovon die Version in wd abgeleitet wurde
      && nNr != nDerivedFrom)   // und diese sich unterscheiden:
  {
    std::cout << "It seems your work is derived from version " << nDerivedFrom << ",\n";
    std::cout << "but you want to create a new version based on " << nNr << ".\n";
    std::cout << "That does not make sense! If you want to do it anyway, delete the file\n";
    std::cout << "  " << conf->attribute("tracking") << "\n";
    std::cout << "and try again.\n";
    return(-1);
  }
  
  // Nun kann man die normale Reihenfolge beachten:
  // nDerivedFrom      nNr
  // ------------------------------------------
  //  -1               -1             nCurrent
  //   != -1           -1             nDerivedFrom
  //  -1              != -1           nNr
  //   != -1          != -1           nNr
  if (nNr == -1)
  {
    if (nDerivedFrom != -1)
      nNr = nDerivedFrom;
    else 
      nNr = nCurrent;      
  }
  
  // Wenn es noch gar keine Version gibt, muss ich halt die erste anlegen
  if (nNr < 0)
  {
    if (cmd == CMD_DIFF)
    {
      // Ein Vergleich ohne Vergleichspartner ist sinnlos
      std::cout << "Error: There isn't anything to compare to.\n";
      return(-1);
    }

    refDir      = "";
    oldNrString = "";
  }
  else
  {
    oldNrString = getVersionNumberString(nNr);
    refDir      = getConfDir(conf, "repository") + oldNrString + "/";
    
    // Prfen, ob es das geforderte Referenzverzeichnis gibt    
    if (existDir(refDir) == false)
    {
      std::cout << "Error: reference directory " << refDir << " does not exist.\n";
      return(-1);
    }    
  }
  
  // Create path to new directory
  newNrString = getVersionNumberString(nCurrent+1);
  targetDir   = getConfDir(conf, "repository") + newNrString;
    
  // Just tell about what's going on
  if (cmd == CMD_DIFF)
  {
    std::cout << "Comparing\n";
    std::cout << "  " << wd << "\n";
    std::cout << "to\n";
    std::cout << "  " << refDir << "\n";
  }
  else
  {
    std::cout << "Saving\n";
    std::cout << "  " << wd << "\n";
    std::cout << "to\n";
    std::cout << "  " << targetDir << "/\n";
    if (refDir.length())
    {
      std::cout << "based on\n";
      std::cout << "  " << refDir << "\n";
    }
  }
    
  if (cmd == CMD_CREATE)
  {
    out = new SimpleXMLTransfer();
    out->setName("SiVeCo");
    out->addAttribute("version", "1");
    
    xml = new SimpleXMLTransfer();
    xml->setName("label");
    
    labelToUse = "none";
    if (wd.compare(getConfDir(conf, "workingDirectory")) != 0)
    {
      labelToUse = wd.substr(0, wd.length()-1);
      while (labelToUse.find('/') != std::string::npos)
        labelToUse = labelToUse.substr(1);
    }      
    if (label.length() != 0)
      labelToUse = label;          
    xml->setContent(labelToUse);
    out->addChild(xml);

    xml = new SimpleXMLTransfer();
    xml->setName("date");
    out->addChild(xml);
    xml->setContent(getDateString());

    xml = new SimpleXMLTransfer();
    xml->setName("time");
    out->addChild(xml);
    xml->setContent(getTimeString());

    // newly created revisions do get system charset of course
    {
      std::string charset = conf->getString("sys.charset.filenames", "");
      if (charset.length())
      {
        xml = new SimpleXMLTransfer();
        xml->setName("charset");
        xml->addAttribute("filenames", charset);
        out->addChild(xml);
      }
    }
    
    
    // Wenn diese Version nicht auf der vorigen basiert und nicht die erste Version
    // ist, muss ich angeben, auf welcher Version sie basiert.
    if (refDir.length() != 0 && nNr != nCurrent)
    {
      xml = new SimpleXMLTransfer();
      xml->setName("basedOn");
      out->addChild(xml);
      xml->setContent(itoStr(nNr, ' ', 1));
    }
    
    xml = new SimpleXMLTransfer();
    xml->setName("description");
    out->addChild(xml);
    
    // Wenn es eine ChangeLog-Datei gibt, nehme ich deren Inhalt
    if (conf->indexOfAttribute("changelog") >= 0 && conf->attribute("changelog").length())
    {
      std::string changelog = wd + conf->attribute("changelog");
      std::string buf;
      int         retCode;
      
      retCode = getChangeLogContent(buf, changelog);
      
      switch (retCode)
      {
       case -1:
        // Es war ein Fehler drin. Damit die Daten nicht verloren gehen und der Benutzer
        // alles sofort bemerkt, breche ich hier ab. Weil dies ein hufiger Fehler ist,
        // ist es wichtig, dass an dieser Stelle noch nichts im Dateisystem getan wurde.
        std::cout << buf << "\n";
        return(-1);
        break;
        
       case 0:
        xml->setContent(buf);
        break;
        
       default:
        xml->setContent("none");
        break;
      }
    }
    
    // Create file
    ofilename = targetDir + ".xml";
    ofile.open(ofilename.c_str());
    if (!ofile)
    {
      std::cout << "Error creating file: " << ofilename << "\n";
      return(-1);
    }    
    
    // Create new directory
    if (mkdir(targetDir.c_str(), S_IRWXU))
    {
      std::cout << "Error when creating new directory\n";
      return(-1);
    }    
  }
  
  // pre-create-command mit dem Label, das die neue Version bekommt.
  if (cmd == CMD_CREATE)
    if (conf->indexOfAttribute("preCreateCmd") >= 0 && conf->attribute("preCreateCmd").length())
      executeCmd(wd + conf->attribute("preCreateCmd"), labelToUse);

  // regex
  compileRegEx();
  
  // recurse through working directory and save things...
  xml = new SimpleXMLTransfer();
  xml->setName("links");
  targetDir += "/";
  
  if (create("", xml))
  {
    std::cout << "Error\n";

    freeRegEx();
    
    return(-1);
  }
  
  if (cmd == CMD_CREATE)
    out->addChild(xml);
  
  // Bestimmen, welche Dateien nicht mehr vorhanden sind
  fillDeletedFileList("");
  
  // --- Infos schreiben --------------------------------------------
  if (conf->getChild("verbosity")->getChild("new")->attributeAsInt("stdout", 0) == 1
      || cmd == CMD_DIFF)
  {
    std::cout << "New files:\n";
    for (unsigned int ucnt=0; ucnt<newFiles.size(); ucnt++)
      std::cout << "  " << newFiles[ucnt] << "\n";
  }  
  if (conf->getChild("verbosity")->getChild("changed")->attributeAsInt("stdout", 0) == 1
      || cmd == CMD_DIFF)
  {
    std::cout << "Changed files:\n";
    for (unsigned int ucnt=0; ucnt<changedFiles.size(); ucnt++)
      std::cout << "  " << changedFiles[ucnt] << "\n";
  }  
  if (conf->getChild("verbosity")->getChild("deleted")->attributeAsInt("stdout", 0) == 1
      || cmd == CMD_DIFF)
  {
    std::cout << "Deleted files:\n";
    for (unsigned int ucnt=0; ucnt<deletedFiles.size(); ucnt++)
      std::cout << "  " << deletedFiles[ucnt] << "\n";
  }  
  if (conf->getChild("verbosity")->getChild("excluded")->attributeAsInt("stdout", 0) == 1)
  {
    std::cout << "Excluded files:\n";
    for (unsigned int ucnt=0; ucnt<excludedFiles.size(); ucnt++)
      std::cout << "  " << excludedFiles[ucnt] << "\n";
  }  

  if (cmd == CMD_CREATE)
  {
    verboseOut = new SimpleXMLTransfer();
    verboseOut->setName("verboseOutput");
  
    if (conf->getChild("verbosity")->getChild("new")->attributeAsInt("file", 0) == 1)
    {
      xml = new SimpleXMLTransfer();
      xml->setName("new");
      content = "\n";
      for (unsigned int ucnt=0; ucnt<newFiles.size(); ucnt++)
        content += "      " + newFiles[ucnt] + "\n";
      content += "    ";
      xml->setContent(content);
      verboseOut->addChild(xml);
    }
    if (conf->getChild("verbosity")->getChild("changed")->attributeAsInt("file", 0) == 1)
    {
      xml = new SimpleXMLTransfer();
      xml->setName("changed");
      content = "\n";
      for (unsigned int ucnt=0; ucnt<changedFiles.size(); ucnt++)
        content += "      " + changedFiles[ucnt] + "\n";
      content += "    ";
      xml->setContent(content);
      verboseOut->addChild(xml);
    }  
    if (conf->getChild("verbosity")->getChild("deleted")->attributeAsInt("file", 0) == 1)
    {
      xml = new SimpleXMLTransfer();
      xml->setName("deleted");
      content = "\n";
      for (unsigned int ucnt=0; ucnt<deletedFiles.size(); ucnt++)
        content += "      " + deletedFiles[ucnt] + "\n";
      content += "    ";
      xml->setContent(content);
      verboseOut->addChild(xml);
    }  
    if (conf->getChild("verbosity")->getChild("excluded")->attributeAsInt("file", 0) == 1)
    {
      xml = new SimpleXMLTransfer();
      xml->setName("excluded");
      content = "\n";
      for (unsigned int ucnt=0; ucnt<excludedFiles.size(); ucnt++)
        content += "      " + excludedFiles[ucnt] + "\n";
      content += "    ";
      xml->setContent(content);
      verboseOut->addChild(xml);
    }
  
    out->addChild(verboseOut);
  
    // out in ofile schreiben
    out->print(ofile);
    ofile.close();
    delete out;
  }
  
  // Beim Diff das changelog anzeigen. blicherweise wird es durch exclFile ausgeschlossen
  // sein. Wenn man dann ein Diff macht, sieht man gar nicht, dass man schon etwas in das
  // changelog geschrieben hat.
  if (cmd == CMD_DIFF)
  {
    std::string changelog = conf->attribute("changelog", "");
    
    if (changelog.length())
    {
      std::string buf;
      int         retCode;
      
      changelog = wd + changelog;
      retCode   = getChangeLogContent(buf, changelog);
      
      std::cout << "ChangeLog:\n";
      
      if (retCode != 1)
        std::cout << buf << "\n";
    
      if (retCode < 0)
        // Es war ein Fehler drin. Damit die Daten nicht verloren gehen und der Benutzer
        // alles sofort bemerkt, breche ich hier ab. Weil dies ein hufiger Fehler ist,
        // ist es wichtig, dass an dieser Stelle noch nichts im Dateisystem getan wurde.
        return(-1);
    }        
  }
  
  // start external diff for changed files?
  if (cmd == CMD_DIFF && type.find("exec") != std::string::npos)
  {
    std::string command;
    
    for (unsigned int ucnt=0; ucnt<changedFiles.size(); ucnt++)
    {
      if (conf->indexOfAttribute("diff") < 0)
        std::cerr << "No attribute \"diff\" in configfile.\n";
      else
      {
        command = conf->attribute("diff") + " \"" 
          + refDir + changedFiles[ucnt] + "\" \""
          + wd     + changedFiles[ucnt] + "\"";
        if (type[0] == 'b')
          command += " &";
        system(command.c_str());
      }
    }
  }
  
  if (cmd == CMD_CREATE)
  {
    // Aufschreiben, dass die Version im workingDir von der gerade erstellten 
    // abhngt.
    if (setDerivedFrom(conf, wd, nCurrent+1))
      return(-1);
    
    // Neue ChangeLog-Datei erstellen
    if (conf->indexOfAttribute("changelog") >= 0 && conf->attribute("changelog").length())
      if (createEmptyChangeLog(wd + conf->attribute("changelog")))
        return(-1);
    
    // after-create-command mit der Nummer, die die neue Version bekommen hat.
    if (conf->indexOfAttribute("afterCreateCmd") >= 0 && conf->attribute("afterCreateCmd").length())
      executeCmd(wd + conf->attribute("afterCreateCmd"), itoStr(nCurrent+1, ' ', 1));
  }
    
  freeRegEx();
  
  return(0);
}

std::string SiVeCoCreate::getOriginalFileRel(std::string path,
                                             std::string refDir)
{
  struct stat     Attribs;              // ein Puffer fr die Dateiattribute
  
  lstat(path.c_str(), &Attribs);
  
  // Wenn path nur ein Link auf eine andere Datei ist,
  // nehme ich diese Datei.
  if (S_ISLNK(Attribs.st_mode) != 0)
  {
    return(getLinkTarget(path));
  }
  else
  {    
    // Angabe relativ machen
    path.replace(0, refDir.length(), "");
  
    std::string tmp1 = path;
    std::string tmp2 = "";
  
    while (tmp1.find("/") != std::string::npos)
    {
      tmp1.replace(0, tmp1.find("/")+1, "");
      tmp2 += "../";    
    }
      
    path = "../" + tmp2 + oldNrString + "/" + path;
    
    return(path);
  }
}

/**
 * Compile regular expression for exclusion of things
 */
int SiVeCoCreate::compileRegEx()
{
  std::string tmp;
  
  tmp = conf->getChild("exclude")->attribute("directories");
  if (tmp.length())
  {
    exclDir = new regex_t;
    if (regcomp(exclDir, tmp.c_str(), REG_EXTENDED))
    {
      std::cout << "Error compiling regular expression for directories: "
        << conf->getChild("exclude")->attribute("directories") << "\n";
      return(-1);
    }
  }
  else
    exclDir = (regex_t*)0;
 
  tmp = conf->getChild("exclude")->attribute("files");
  if (tmp.length())
  {
    exclFile = new regex_t;
    if (regcomp(exclFile, tmp.c_str(), REG_EXTENDED))
    {
      std::cout << "Error compiling regular expression for files: " 
        << conf->getChild("exclude")->attribute("files") << "\n";
      return(-1);
    }
  }
  else
    exclFile = (regex_t*)0;

  return(0);
}

void SiVeCoCreate::freeRegEx()
{
  if (exclDir != (regex_t*)0)
    delete exclDir;
  if (exclFile != (regex_t*)0)
    delete exclFile;
}
