/*
 * Copyright (C) 2005-2006  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "filetools.h"

#include <fstream>
#include <iostream>
#include <sys/types.h> 
#include <sys/stat.h> 
#include <unistd.h> 
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef WIN32
# include <direct.h>   // mkdir
#endif




std::string FileTools::createAbsPath(std::string base,
                                     std::string rel)
{
  std::string            path = base + rel;
  std::string            path1;
  std::string::size_type pos1;
  std::string::size_type pos2;

  // Reduce /./ to /
  while ((pos1 = path.find("/./")) != std::string::npos)
  {
    path1 = path.substr(0, pos1);
    path  = path.substr(pos1+2);
    path  = path1 + path;
  }
  
  while ((pos1 = path.find("../")) != std::string::npos)
  {
    // std::cout << path << "\n";
    path1 = path.substr(0, pos1-1);
    path  = path.substr(pos1+2);
    
    if ((pos2 = path1.rfind('/')) != std::string::npos)
    {
      path1 = path1.substr(0, pos2);
      path  = path1 + path;
    }
  }
    
  return(path);
}

std::string FileTools::createRelPath(std::string refPath,
                                     std::string absPath)
{
  if (absPath.find(refPath) == 0)
  {
    // Der einfachste Fall
    return("./" + absPath.substr(refPath.length()));
  }
  else
  {
    std::string            basis = refPath.substr(0, refPath.rfind('/'));
    std::string            hoch  = "";
    std::string::size_type pos;

    // Den einen so lange hinten abschneiden, bis er in dem anderen vorkommt.
    while (absPath.find(basis) == std::string::npos && (pos = basis.rfind('/')) != std::string::npos)
    {
      basis = basis.substr(0, pos);
      hoch.append("../");
    }
    
    // 'Basis' ist jetzt die gemeinsame Basis, 'hoch' enthält den Weg
    // vom Referenzpfad zur gemeinsamen Basis.
          
    return("./" + hoch + absPath.substr(basis.length()+1));
  }  
}

bool FileTools::fileExists(std::string path)
{
  std::ifstream testDat;
  
  // std::cout << "bool FileTools::fileExists(std::string path = " << path << ")\n";
  
  testDat.open((const char*)path.c_str());

  if (!testDat) 
    return(false);
  else 
  { 
    testDat.close();
    return(true);
  }  
}

void FileTools::openAndCreateBackup(std::ofstream& out, std::string path)
{
  if (fileExists(path))
  {
    copy(path + "_backup", path);
  }
  
  out.open(path.c_str());
}

int FileTools::isFileUpToDate(std::string ref, std::string result)
{
  if (fileExists(result) == false)
    return(-2);
  else
  {
    // take remaining data from file
    struct stat     AttribsRef;
    struct stat     AttribsRes;
    
    stat(ref.c_str(),    &AttribsRef);
    stat(result.c_str(), &AttribsRes);

    if (AttribsRes.st_mtime > AttribsRef.st_mtime)
      return(1);
    else        
      return(-1);
  }
}

int FileTools::makeSureDirExists(std::string dirname)
{
  int nRetCode = 0;
  
  // std::cout << "int FileTools::makeSureDirExists(" << dirname << ")\n";
  
  // I expect an absolute path beginning with '/'
  if (dirname.length() < 2)
    return(-1);
  
  std::string::size_type posstart = 2;
  std::string::size_type pos;
  
  while (nRetCode == 0 && 
         posstart < dirname.length() && 
         (pos = dirname.find('/', posstart)) != std::string::npos)
  {
#ifdef WIN32
    mkdir(dirname.substr(0, pos).c_str());
#else
    if (mkdir(dirname.substr(0, pos).c_str(), 0777))
    {
      if (errno != EEXIST)
      {
        int errsv = errno;
        std::cout << "mkdir " << dirname.substr(0, pos) << "\n";
        nRetCode = -1;
        std::cout << "Error on mkdir: " << strerror(errsv) << "\n";
      }
    }
#endif    
    posstart = pos+1;
  }
  
  if (nRetCode == 0 && dirname[dirname.length()-1] != '/')
  {
#ifdef WIN32
    mkdir(dirname.c_str());
#else
    if (mkdir(dirname.c_str(), 0777))
    {
      if (errno != EEXIST)
      {
        int errsv = errno;
        std::cout << "mkdir " << dirname << "\n";
        nRetCode = -1;
        std::cout << "Error on mkdir: " << strerror(errsv) << "\n";
      }
    }
#endif
  }
          
  return(nRetCode);
}

std::string FileTools::getHomeDir(std::string appdirname)
{
  std::string tmp;
  
#ifdef WIN32
  tmp   = getenv("USERPROFILE");
  tmp  += "/";
  tmp  += appdirname;
#else
  tmp   = getenv("HOME");
  tmp  += "/.";
  tmp  += appdirname;
#endif      

  // Gibt es das Verzeichnis?
  struct stat     Attribs;              // ein Puffer für die Dateiattribute
  
  if (stat(tmp.c_str(), &Attribs))
  {
    // Das Verzeichnis muss angelegt werden
#ifdef WIN32
    mkdir(tmp.c_str());
#else
    mkdir(tmp.c_str(), S_IRWXU);
#endif
    
    // Ist es jetzt vorhanden?
    if (stat(tmp.c_str(), &Attribs))
    {
      // dann wird das aktuelle Verzeichnis benutzt
      tmp = "";
    }
    else
    {
      tmp += "/";
    }
  }
  else
  {
    tmp += "/";
  }
  
  // printf("DaLoMisc::getHomeDir() = %s\n", tmp.c_str());
  
  return(tmp);
}

int FileTools::rm(std::string filename)
{
#ifdef WIN32
  // todo
#else
  // linux: man 3 remove
  return(remove(filename.c_str()));
#endif
}

int FileTools::getSize(std::string filename)
{
  struct stat     Attribs;              // ein Puffer für die Dateiattribute
  
  if (stat(filename.c_str(), &Attribs))
  {
    return(-1);
  }
  
  return(Attribs.st_size);
}

std::string FileTools::name(std::string pathAndName)
{
  std::string::size_type pos = pathAndName.rfind('/');
  
  if (pos != std::string::npos)
  {
    return(pathAndName.substr(pos+1));
  }
  else
    return(pathAndName);
}

std::string FileTools::path(std::string pathAndName)
{
  std::string::size_type pos = pathAndName.rfind('/');
  
  if (pos != std::string::npos)
  {
    return(pathAndName.substr(0, pos+1));
  }
  else
    return("");
}

std::string FileTools::removeSuffix(std::string name)
{
  std::string::size_type pos = name.find('.');
  
  if (pos != std::string::npos)
    return(name.substr(0, pos));
  else
    return(name);
}

int FileTools::copy(std::string dest, std::string src)
{
  if (dest[dest.length()-1] == '/')
    dest += name(src);
  
  std::ifstream in;
  std::ofstream out;
  char buf[256];
  
  in.open(src.c_str(), std::ios::binary);
  if (!in)
  {
    return(-1);
  }
  
  out.open(dest.c_str(), std::ios::binary);
  if (!out)
  {
    in.close();
    return(-1);
  }

  while (in)
  {
    in.read(buf, 256);
    out.write(buf, in.gcount());
  }
    
  in.close();
  out.close();
  
  return(0);
}

int FileTools::symlink(std::string dest, std::string src)
{  
#ifdef WIN32
  return(-1);
#else
  return(::symlink(src.c_str(), dest.c_str()));
#endif
}

std::string FileTools::stripAbsFromPath(std::string in)
{
#ifdef WIN32
  // todo
#else
  while (in.length() && in[0] == '/')
    in = in.substr(1);
#endif
  
  return(in);
}

std::string FileTools::cleanString(std::string pathAndName, std::string allowed)
{
  int len = pathAndName.length();
  
  for (int cnt=0; cnt<len; cnt++)
  {
    if (isalnum(pathAndName[cnt]) == 0 &&
        pathAndName[cnt] != '_' &&
        allowed.find(pathAndName[cnt]) == std::string::npos)
    {
      pathAndName[cnt] = '_';
    }        
  }
  
  return(pathAndName);
}

bool FileTools::isDir(std::string name)
{
  struct stat     AttribsRef;
    
  stat(name.c_str(),    &AttribsRef);
   
  return(S_ISDIR(AttribsRef.st_mode) != 0);
}

#if (TEMP_TEXTFILE_VARIANT == 0)
TemporaryTextFile::TemporaryTextFile(std::string name_template)
{
  name_template += "XXXXXX";
  
  char* buf = (char*)malloc(name_template.length()+1);
  strcpy(buf, name_template.c_str());
  
  filedescr = mkstemp(buf);
  if (filedescr != -1)
  {
    fchmod(filedescr, S_IWUSR | S_IRUSR);
    FilenameAndPath = buf;
    fOK = true;
  }
  else
  {
    fOK = false;
  }        
  free(buf);
}
#endif

#if (TEMP_TEXTFILE_VARIANT == 1)
TemporaryTextFile::TemporaryTextFile(std::string name_template)
{
  char         buf[10];
  const int    cMaxCount = 1000000;
  int          cnt       = 0;
  std::string  name;
  
  do 
  {
    sprintf(buf, "%i", cnt);
    name = name_template + buf;
    cnt++;
  } while (FileTools::fileExists(name) && cnt < cMaxCount);
  
  if (cnt == cMaxCount)
    fOK = false;
  else
  {
    outfile.open(name.c_str());
    if (outfile)
    {
      FilenameAndPath = name;      
      fOK = true;
    }
    else
      fOK = false;
  }  
}
#endif

#if (TEMP_TEXTFILE_VARIANT == 2)
TemporaryTextFile::TemporaryTextFile()
{
  tmpfiledescr = tmpfile();
  if (tmpfiledescr != NULL)
  {
    // big trouble: what is the filename?
    FilenameAndPath = "";    
    fOK = true;    
  }
  else
  {
    fOK = false;
  }
}
#endif

void TemporaryTextFile::write(std::string line)
{
#if (TEMP_TEXTFILE_VARIANT == 0)
  ::write(filedescr, line.c_str(), line.length());
#endif
  
#if (TEMP_TEXTFILE_VARIANT == 1)
  outfile << line.c_str();
#endif
  
#if (TEMP_TEXTFILE_VARIANT == 2)
  std::string  tmpstr;
  unsigned int uLen;
  
  // Work around potential security hole (see man fprintf):
  uLen = tmpstr.length();
  for (unsigned int uCnt=0; uCnt<uLen; uCnt++)
    if (tmpstr[uCnt] == '%')
      tmpstr[uCnt] = ' ';
  fprintf(tmpfiledescr, "%s", tmpstr.c_str());
#endif
}

void TemporaryTextFile::close()
{
#if (TEMP_TEXTFILE_VARIANT == 0)
  fsync(filedescr);
  ::close(filedescr);
#endif
  
#if (TEMP_TEXTFILE_VARIANT == 1)
  outfile.close();
#endif
  
#if (TEMP_TEXTFILE_VARIANT == 2)
  // File would be deleted on close. Something like fsync?
#endif
}

std::string TemporaryTextFile::GetGlobalTempDirPath()
{
#ifdef WIN32
  return("C:/temp/");
#else
  return("/tmp/");
#endif
}
