#include "info.h"

#include <iostream>

#include "tools.h"
#include "XMLExtract.h"

/**
 * 
 */
int printHistory(SimpleXMLTransfer* conf,
                 int                nNr,
                 bool               fHTML)
{
  // Die beschreibende Datei ffnen
  std::string        infilename;
  SimpleXMLTransfer* logfile;
  SimpleXMLTransfer* item;
  
  std::string        basedOnStr;
  std::string        labelStr;
  std::string        dateStr;
  std::string        timeStr;
  std::string        descrStr;
  
  
  infilename = getConfDir(conf, "repository") + getVersionNumberString(nNr) + ".xml";
  if (existDat(infilename) == false)
    return(0); // es ist kein Fehler, wenn es eine Version nicht gibt
  
  try
  {
    logfile = new SimpleXMLTransfer(infilename);
    
    // Die Version berprfen
    if (canHandleFileVersion(logfile->attributeAsInt("version")) == false)
    {
      std::cout << "Can't handle that version!\n";
      delete logfile;
      return(-1);      
    }
  
    // Daten auslesen
    try
    {
      item       = logfile->getChild("basedOn");
      basedOnStr = item->getContentString();
    }
    catch (XMLException e)
    {
      basedOnStr = "";
    }
            
    labelStr = logfile->getChild("label")->getContentString();
    dateStr  = logfile->getChild("date")->getContentString();
    timeStr  = logfile->getChild("time")->getContentString();
    
    if (XMLExtract(descrStr, infilename, "SiVeCo.description", fHTML))
    {
      std::cout << descrStr << "\n";
      return(-1);      
    }
      
    // Daten anzeigen
    if (fHTML)
    {
      std::cout << "<table width=\"100%\"><tr><td>";
      std::cout << "(" << nNr;
      if (basedOnStr.length() != 0)
        std::cout << " based on " << basedOnStr;
      std::cout << ")</td><td>";
      
      std::cout << labelStr << "</td><td>" << dateStr << " " << timeStr << "</td></tr></table>\n";
      std::cout << descrStr << "\n";
      std::cout << "<hr width=\"100%\">\n";
    }
    else
    {
      std::cout << "(" << nNr;
      if (basedOnStr.length() != 0)
        std::cout << " based on " << basedOnStr;
      std::cout << ")                  ";
      
      std::cout << labelStr << "             " << dateStr << " " << timeStr << "\n\n";
      std::cout << descrStr << "\n\n";
      std::cout << "     ------------------------------------------------------------\n\n";
    }
  }
  catch (XMLException e)
  {
    std::cout << e.what() << "\n";
    return(-1);
  }
  
  return(0);  
}

/**
 * 
 */
int printHistory(SimpleXMLTransfer* conf,
                 std::string        type)
{
  bool fHTML           = false;
  bool fHTMLHeader     = false;
  
  // Laufrichtung und -weite festlegen
  int maxnum = getCurrentVersionNumberInt(conf);
  int start;
  int end;
  int step;
  
  if (type.find("rev") != std::string::npos)
  {
    start = maxnum;
    end   = 0;
    step  = -1;
  }
  else
  {
    start = 0;
    end   = maxnum;
    step  = 1;
  }

  // Ausgabeformat
  if (type.find("htm") != std::string::npos)
  {
    fHTML = true;  
    
    if (type.find("NH") == std::string::npos)
      fHTMLHeader = true;
  }
  
  if (fHTMLHeader)
    std::cout << "<html><body>\n";
  
  // Alle ausgeben
  end += step;
  for (int cnt=start; cnt != end; cnt += step)
    if (printHistory(conf, cnt, fHTML))
      return(-1);
  
  if (fHTMLHeader)
    std::cout << "</body></html>\n";
  
  return(0);
}

/**
 * Print information about a version
 */
int printInfo(SimpleXMLTransfer* conf,
              int                nNr,
              std::string        type)
{
  // Die beschreibende Datei ffnen
  std::string        infilename;
  SimpleXMLTransfer* logfile;
  std::string        label;
  
  if (nNr == -1)
    nNr = getCurrentVersionNumberInt(conf);
    
  infilename = getConfDir(conf, "repository") + getVersionNumberString(nNr) + ".xml";
  if (existDat(infilename) == false)
  {
    std::cout << "Error: repository file " << infilename << " does not exist.\n";
    return(-1);
  }
  try
  {
    logfile = new SimpleXMLTransfer(infilename);
    // Die Version berprfen
    if (canHandleFileVersion(logfile->attributeAsInt("version")) == false)
    {
      std::cout << "Can't handle that version!\n";
      delete logfile;
      return(-1);      
    }
    // label auslesen
    label = logfile->getChild("label")->getContentString();

    if (type.compare("changes") == 0)
    {
      std::string descrStr;
      XMLExtract(descrStr, infilename, "SiVeCo.description", true);
      std::cout << descrStr << "\n";
    }
    else
    {
      // Infos anzeigen
      if (type.compare("nameLabel") == 0)
        std::cout << conf->attribute("name") << "-";    
      std::cout << label << "\n";
    }
  }
  catch (XMLException e)
  {
    std::cout << e.what() << "\n";
    return(-1);
  }
  
  return(0);
}
