/*
 * Copyright (C) 2003-2007  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "lib_conversions.h"
#include <ctype.h>
#include <math.h>

#define DEBUG 0

#include <stdio.h>

std::string doubleToString(double dVal)
{
  char tmp[30];
  int  nStart = 0;

  sprintf(tmp, "%2.8g", dVal);
  while (tmp[nStart] == ' ')
    nStart++;
  
  return(&(tmp[nStart]));
}

std::string itoStr(long int nr, char fill, unsigned char digits, bool prettyneg)
{
  std::string  tmp = "";
  char    digit;
  bool    neg = false;
  int     Cnt;
  
  if (nr < 0) 
    {
      neg = true;
      nr *= -1;
    }

  
  if (nr == 0)
    {
      for (int Cnt = 0; Cnt < digits-1; Cnt++) tmp += fill;
      tmp += '0';
    }
  else 
    {
      Cnt = 0;
      while (Cnt < digits || nr > 0 || neg)
        {
          if (nr > 0) 
            digit = (nr % 10)+48;
          else 
	    {
	      if (neg) 
		{
		  if (prettyneg) digit = '-';
		  else 
		    {
		      tmp = '-' + tmp;
		      digit = fill;
		    }
		  neg = false;
		}
	      else digit = fill;
	    }
	    
          tmp = digit + tmp;
          nr /= 10;
	  Cnt++;
        }
    }
  
  return(tmp);
}

std::string itoStr(long int nr, int nDecPtAt, bool fShorten, unsigned char digitspre)
{
  bool fNeg = false;
  if (nr < 0)
  {
    fNeg = true;
    nr   = -nr;
  }
  
  std::string tmp = itoStr(nr, '0', nDecPtAt+1);

  if (nDecPtAt)
  {
    int nDecPtAtLoc = nDecPtAt;
    // In einem String steht rechtsbündig eine Zahl, in die ein Dezimalpunkt
    // eingefügt werden soll.
    if (fShorten)
    {
      // Die Stellen, die rechts vom Dezimalpunkt Null sind, kann man löschen.
      int s=0;
      while (tmp[tmp.length()-1-s] == '0' && nDecPtAtLoc > 1)
      {
        nDecPtAtLoc--;
        s++;
      }
      tmp = tmp.substr(0, tmp.length()-s);
    }
    
    int len = tmp.length();
    int src = len-nDecPtAtLoc;
    
    tmp += " ";
    for (int n=len+1; n>src; n--)
    {
      if (tmp[n-1] == ' ')
        tmp[n] = '0';
      else
        tmp[n] = tmp[n-1];
    }
    tmp[src] = '.';
  }
  
  if (fNeg)
    tmp = "-" + tmp;
  
  return(tmp);
}

long int mypow10(int exponent)
  // ich will int als output, nicht double
{
  int tmp = 1;

  for (int Cnt=0; Cnt<exponent; Cnt++) tmp *= 10;
  
  return(tmp);
}

std::string ftoStr(double        nr, 
                   unsigned char digitspre, 
                   unsigned char digitspost, 
                   bool          prettyneg,
                   bool          fComma)
{
  std::string   tmp;
  int      pre;
  int      post;
  bool     neg = false;


  if (nr < 0) 
    {
      neg = true;
      nr *= -1;
    }
  
  pre = int(floor(nr));
  // Nachkommastellen werden gerundet
  post = int(floor(   (nr - pre) * mypow10(digitspost)  +0.5 ));
  if (post == mypow10(digitspost))
  {
    post = 0;
    pre++;
  }
  
  tmp = itoStr(pre, ' ', digitspre);
  if (fComma)
    tmp += ",";
  else
    tmp += "."; 
  tmp += itoStr(post, '0', digitspost);

  
  if (neg)
    {
      int pos;

      pos = tmp.length()-1;
      while ((tmp[pos] != ' ') && (pos > 0)) pos--;

      if (pos == 0 && tmp[pos] != ' ') tmp = ' ' + tmp; // nur für den Notfall
      
      if (prettyneg) tmp[pos] = '-';
      else tmp = tmp.substr(0, pos+1) + '-' + tmp.substr(pos+1, tmp.length()-pos);      
    }

  return(tmp);
}

int hex2int(std::string hexIn)
{
  int tmp = 0;
  int wert1 = 1;
  int wert2;
  
  for (unsigned int n= 0; n< hexIn.length(); n++)
    {
      wert2 = hexIn[hexIn.length()-n-1];
      if (wert2 >=97 ) wert2-= 87;
      if (wert2 >=65 ) wert2-= 55;      
      if (wert2 >= 48) wert2-= 48;            
      tmp+= wert2 * wert1;
      wert1*= 16;
    }
  return(tmp);
}

std::string strU(std::string in) 
{
  std::string tmp = "";
  
  for (unsigned int uCnt = 0; uCnt < in.length(); uCnt++)
    {
      tmp += toupper(in[uCnt]); 
    } 
  return(tmp); 
}

/**
 * Removes leading and trailing whitespace
 */
std::string trim(std::string in)
{
  int nL  = in.length();
  int nCnt;
  
  // Die Anzahl der führenden Zeichen ermitteln
  for (nCnt=0; nCnt < nL && isspace(in[nCnt]); nCnt++)
    {
    }
    
  // Die führenden Zeichen löschen
  in.replace(0, nCnt, "");
  
  nL = in.length();
  if (nL > 0)
    {
      // Die Anzahl der abschliessenden Zeichen ermitteln
      for (nCnt = nL-1; nCnt > 0 && isspace(in[nCnt]); nCnt--)
        {
        }
      // Die abschliessenden Zeichen löschen
      in.replace(nCnt+1, nL-nCnt, "");     
    }
  
  return(in);
}

std::string itoHexStr(long unsigned int nr, int len)
{
  std::string tmp = "";
  char        ch;

  while (nr > 0 || len > 0)
  {  
    ch = 48 + (nr & 0x0F);
    if (ch > 57) 
      ch += 7;
    tmp = ch + tmp;
    
    nr >>= 4;
    len -= 4;
  }

  return(tmp);  
}

void float2MulShift(float fl, int& mul, int& shift)
{
  int          nMulBits = 15;
  long long    lMaxNum  = (1l << nMulBits)-1;
  long long    lDiv;
  
  // Dieses sollte auf jeden Fall möglich sein:
  mul   = (long long)(fl);
  shift = 0;
       
#if DEBUG == 1
  printf("lMaxNum = %lli, fl=%f\n", lMaxNum, fl);
#endif
  
  // Nun was gescheites berechnen:
  lDiv = (long long)(lMaxNum / fl);
              
#if DEBUG == 1
  printf("lDiv = %lli\n", lDiv);
#endif
  
  while ((1LL<<shift) < lDiv)
  {
    shift++;
  }
  shift--;
  
#if DEBUG == 1
  printf("shift = %i\n", shift);
#endif
  
  mul = (int)floor(  (fl * (1LL<<shift)) +0.5 );
       
#if DEBUG == 1
  printf("mul = %i\n", mul);
#endif
  
  // Solange mul eine gerade Zahl ist kann man sie und den Schiebefaktor
  // halbieren ohne das Ergebnis ungenauer zu machen.
  while ((mul & 1) == 0 && shift > 0)
  {
    mul >>= 1;
    shift--;
  }
}

std::string strWithMinLen(std::string in, std::string::size_type nMin)
{
  if (in.length() >= nMin)
    return(in);
  else
  {
    return(in + std::string(nMin - in.length(), ' ')       );
  }
}

void lib_conversions_tests()
{
  printf("Keine Nachkommastellen\n");
  printf("%s\n", itoStr(0, 0, false).c_str());
  printf("%s\n", itoStr(5, 0, false).c_str());
  printf("%s\n", itoStr(10, 0, false).c_str());
  printf("%s\n", itoStr(15, 0, false).c_str());
  printf("%s\n", itoStr(100, 0, false).c_str());
  printf("%s\n", itoStr(150, 0, false).c_str());
  
  printf("Eine Nachkommastelle\n");
  printf("%s\n", itoStr(0, 1, false).c_str());
  printf("%s\n", itoStr(5, 1, false).c_str());
  printf("%s\n", itoStr(10, 1, false).c_str());
  printf("%s\n", itoStr(15, 1, false).c_str());
  printf("%s\n", itoStr(100, 1, false).c_str());
  printf("%s\n", itoStr(150, 1, false).c_str());

  printf("Zwei Nachkommastellen\n");
  printf("%s\n", itoStr(0, 2, false).c_str());
  printf("%s\n", itoStr(5, 2, false).c_str());
  printf("%s\n", itoStr(10, 2, false).c_str());
  printf("%s\n", itoStr(15, 2, false).c_str());
  printf("%s\n", itoStr(100, 2, false).c_str());
  printf("%s\n", itoStr(150, 2, false).c_str());

  printf("Drei Nachkommastellen\n");
  printf("%s\n", itoStr(0, 3, false).c_str());
  printf("%s\n", itoStr(5, 3, false).c_str());
  printf("%s\n", itoStr(10, 3, false).c_str());
  printf("%s\n", itoStr(15, 3, false).c_str());
  printf("%s\n", itoStr(100, 3, false).c_str());
  printf("%s\n", itoStr(150, 3, false).c_str());

  
  printf("Keine Nachkommastellen\n");
  printf("%s\n", itoStr(0, 0).c_str());
  printf("%s\n", itoStr(5, 0).c_str());
  printf("%s\n", itoStr(10, 0).c_str());
  printf("%s\n", itoStr(15, 0).c_str());
  printf("%s\n", itoStr(100, 0).c_str());
  printf("%s\n", itoStr(150, 0).c_str());
  
  printf("Eine Nachkommastelle\n");
  printf("%s\n", itoStr(0, 1).c_str());
  printf("%s\n", itoStr(5, 1).c_str());
  printf("%s\n", itoStr(10, 1).c_str());
  printf("%s\n", itoStr(15, 1).c_str());
  printf("%s\n", itoStr(100, 1).c_str());
  printf("%s\n", itoStr(150, 1).c_str());

  printf("Zwei Nachkommastellen\n");
  printf("%s\n", itoStr(0, 2).c_str());
  printf("%s\n", itoStr(5, 2).c_str());
  printf("%s\n", itoStr(10, 2).c_str());
  printf("%s\n", itoStr(15, 2).c_str());
  printf("%s\n", itoStr(100, 2).c_str());
  printf("%s\n", itoStr(150, 2).c_str());

  printf("Drei Nachkommastellen\n");
  printf("%s\n", itoStr(0, 3).c_str());
  printf("%s\n", itoStr(5, 3).c_str());
  printf("%s\n", itoStr(10, 3).c_str());
  printf("%s\n", itoStr(15, 3).c_str());
  printf("%s\n", itoStr(100, 3).c_str());
  printf("%s\n", itoStr(150, 3).c_str());
  
  printf("Weitere\n");
  printf("-0.004 %s\n", itoStr(-4, 3).c_str());  
}
