/*
 * Copyright (C) 2007 Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "move.h"

#include <unistd.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <iostream>
#include <string>
#include <dirent.h>     // fr die Verzeichnisoperationen
#include <sys/stat.h>   // Abfragen der Dateiattribute
#include <sys/types.h>

#include "tools.h"

SiVeCoMove::SiVeCoMove(std::string        confpath, 
                       SimpleXMLTransfer* ptr_conf)
{
  configpath = confpath;
  conf       = ptr_conf;  
}

int SiVeCoMove::run(int num, int nmaxnum)
{
  std::string        infilename;
  SimpleXMLTransfer* logfile;

  int                nBasedOn;
  std::string        basedOnStr;
  std::string        labelStr;
  std::string        dateStr;
  std::string        timeStr;
  std::string        tag;
  
  infilename = getConfDir(conf, "repository") + getVersionNumberString(num) + ".xml";
  if (existDat(infilename) == false)
    return(0); // es ist kein Fehler, wenn es eine Version nicht gibt
  
  try
  {
    logfile = new SimpleXMLTransfer(infilename);
    
    // Die Version berprfen
    if (canHandleFileVersion(logfile->attributeAsInt("version")) == false)
    {
      std::cout << "Can't handle that version!\n";
      delete logfile;
      return(-1);      
    }

    // Daten auslesen
    {
      int idx = logfile->indexOfChild("basedOn");
      if (idx >= 0)
        nBasedOn = logfile->getChildAt(idx)->getContentInt();
      else
        nBasedOn = -1;
    }
    if (nBasedOn < 0)
      basedOnStr = "";
    else
      basedOnStr = itoStr(nBasedOn, ' ', 1);
            
    labelStr = logfile->getChild("label")->getContentString();
    dateStr  = logfile->getChild("date")->getContentString();
    timeStr  = logfile->getChild("time")->getContentString();
    
    if (dateStr.length() == 10)
    {
      if (dateStr[2] == '.' && dateStr[5] == '.')
      {
        std::string tmp = dateStr;
        dateStr[0] = tmp[6];
        dateStr[1] = tmp[7];
        dateStr[2] = tmp[8];
        dateStr[3] = tmp[9];
        dateStr[4] = '-';
        dateStr[5] = tmp[3];
        dateStr[6] = tmp[4];
        dateStr[7] = '-';
        dateStr[8] = tmp[0];
        dateStr[9] = tmp[1];
      }
    }
    else
      dateStr = "1970-01-01";
    
    if (timeStr.length() < 8)
      timeStr = "12:00:00";
    
    timeStr[5] = ':';
    
    tag = labelStr;
    if (tag.length())
      tag += " ";
    tag += "(" + itoStr(num, ' ', 1) + ")";
    
    // std::cout << "echo --------------- " << num << "/" << nmaxnum << " ------------------------\n";
    if (nBasedOn >= 0)
    {
      std::cout << "BASEDON=\"${REV" << nBasedOn <<"}\"\n";
      std::cout << "custom_set_base\n";
    }
    
    std::cout << "if siveco g " << configpath << " -q -n " << num << " -d ${TMPDIR} -t archive\n";
    std::cout << "  then\n";
    std::cout << "    true\n";
    std::cout << "  else\n";
    std::cout << "    exit -1\n";
    std::cout << "fi\n";
    std::cout << "TAG=\"" << tag << "\"\n";
    std::cout << "LABEL=\"" << labelStr << "\"\n";
    std::cout << "DATE=\"" << dateStr << " " << timeStr << "\"\n";
    std::cout << "siveco i " << configpath << " -n " << num << " -t changes > ${MSGFILE}\n";
    std::cout << "custom_commit\n";
    std::cout << "REV" << num <<"=`custom_get_revid`\n";
    std::cout << "CURREV=`custom_get_revid`\n";
    std::cout << "custom_tag\n";
    std::cout << "custom_clean_files\n";
    // Dateien und Verzeichnisse entfernen
    if (num < nmaxnum)
    {
      DIR*            Verz;                 // der VerzeichnisStrom
      dirent*         VerzEintr;            // hier werden die Daten aus obigem gegeben
      struct stat     Attribs;              // ein Puffer fr die Dateiattribute
      
      std::string rd = getConfDir(conf, "repository") + getVersionNumberString(num);
      
      Verz = opendir(rd.c_str());
      if (Verz != 0)
      {
        while ((VerzEintr = readdir(Verz)) != NULL)
        {
          std::string currentFile = rd + "/" + VerzEintr->d_name;
          lstat(currentFile.c_str(), &Attribs);      
          if (S_ISLNK(Attribs.st_mode) != 0)
          {
            std::cout << "rm \"" << VerzEintr->d_name << "\"\n";
          }
          else
          {
            if (S_ISDIR(Attribs.st_mode) != 0)
            {
              if (strcmp("..", VerzEintr->d_name) != 0 && strcmp(".", VerzEintr->d_name) != 0)
              {
                std::cout << "rm -r \"" << VerzEintr->d_name << "\"\n";
              }
            }
            else
            {
              std::cout << "rm \"" << VerzEintr->d_name << "\"\n";
            }
          }
        }
        closedir(Verz);
      }
    }
    std::cout << "sleep 1\n\n";
  }
  catch (XMLException e)
  {
    std::cout << e.what() << "\n";
    return(-1);
  }
  
  return(0);
}

int SiVeCoMove::run()
{
  std::cout << "#! /bin/sh\n\n";
  
  int maxnum = getCurrentVersionNumberInt(conf);
  for (int n=0; n<=maxnum; n++)
  {
    if (run(n, maxnum))
      return(-1);
  }

  return(0);  
}
