/*
 * Copyright (C) 2004  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "remove.h"

#include <unistd.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <string>
#include <string.h>
#include <ctype.h>
#include <dirent.h>     // fr die Verzeichnisoperationen
#include <sys/stat.h>   // Abfragen der Dateiattribute
#include <sys/types.h>

#include "tools.h"

#define DEBUGFIND 0
#define DODELETE  1

/**
 * 
 */
SiVeCoRemove::SiVeCoRemove(SimpleXMLTransfer* ptr_conf)
{
  conf       = ptr_conf;
  nBaseChars = getConfDir(conf, "repository").length();
}

/**
 * Remove version(s)
 */
int SiVeCoRemove::run(int nNrStart,
                      int nNrEnd)
{
  for (int nNr=nNrStart; nNr<=nNrEnd; nNr++)
  {
    // Beschreibungsdatei lschen
    std::string filename = getConfDir(conf, "repository") + getVersionNumberString(nNr) + ".xml";
    if (existDat(filename))
    {
#if DODELETE == 1
      if (remove(filename.c_str()) != 0)
      {
        std::cout << "Error: unable to remove file " << filename << "\n";
        return(-1);
      }
#else      
      std::cout << "Removing " << filename << "\n";
#endif
    }
    
    // Alle Links lschen, wirkliche Dateien in Liste merken
    if (schrittEins(getConfDir(conf, "repository") + getVersionNumberString(nNr) + "/"))
      return(-1);
  }

  // Liste anzeigen
#if DEBUGFIND == 1
  showList();
#endif
  
  // zu lschende Verzeichnisse durchgehen: Links untersuchen
  {
    DIR*            Verz;                 // der VerzeichnisStrom
    dirent*         VerzEintr;            // hier werden die Daten aus obigem gegeben
    struct stat     Attribs;              // ein Puffer fr die Dateiattribute
    int             len;
    
    Verz = opendir(getConfDir(conf, "repository").c_str());
    if (Verz != 0)
    {
      while ((VerzEintr = readdir(Verz)) != NULL)
      {
        std::string currentFile = getConfDir(conf, "repository");
        currentFile += VerzEintr->d_name;
        stat(currentFile.c_str(), &Attribs);
        
        // Ist es ein Versionsverzeichnis?
        len = strlen(VerzEintr->d_name);
        if (S_ISDIR(Attribs.st_mode) != 0       // Verzeichnis?
            && len == 5)                        // Namenslnge?            
        {          
          // Alle Buchstaben mssen 0-9 sein
          bool fOK = true;
          
          for (int cnt=0; cnt<len; cnt++)          
            if (isdigit((VerzEintr->d_name)[cnt]) == 0)
              fOK = false;
          
          if (fOK && atoi(VerzEintr->d_name) > nNrEnd)
            schrittZwei(currentFile + "/");
        }
      }
      closedir(Verz);      
    }
  }
  
  // Liste anzeigen
#if DEBUGFIND == 1
  showList();
#endif
  
  // Alle Dateien der Liste lschen, bei denen 'fUsed' noch immer false ist
  for (unsigned int ucnt=0; ucnt<files.size(); ucnt++)
    if (files[ucnt].fUsed == false)
    {
      std::string filename = getConfDir(conf, "repository") + files[ucnt].name;
#if DODELETE == 1
      if (remove(filename.c_str()) != 0)
      {
        std::cout << "Error: unable to remove file " << filename << "\n";
        return(-1);
      }
#else
      std::cout << "Removing " << filename << "\n";
#endif
    }

  // zu lschende Verzeichnisse durchgehen: leere lschen
  for (int nNr=nNrStart; nNr<=nNrEnd; nNr++)
  {
    // Beschreibungsdatei lschen
    std::string dirname = getConfDir(conf, "repository") + getVersionNumberString(nNr) + "/";

    if (existDir(dirname))
    {
      if (removeEmptyDirs(dirname))
      {
#if DODELETE == 1
        if (rmdir(dirname.c_str()))
        {
          std::cout << "Error removing directory " << dirname << "\n";
        }
#else
        std::cout << "Removing directory " << dirname << "\n";
#endif
      }
      else
      {
        std::cout << "Not removing directory\n  " << dirname << "\nas its content is still used.\n";
      }
    }
  }
  
  return(0);
}

bool SiVeCoRemove::removeEmptyDirs(std::string currentDir)
{
  DIR*            Verz;                 // der VerzeichnisStrom
  dirent*         VerzEintr;            // hier werden die Daten aus obigem gegeben
  struct stat     Attribs;              // ein Puffer fr die Dateiattribute
  bool            fEmpty = true;

  Verz = opendir(currentDir.c_str());
  if (Verz != 0)
  {
    while ((VerzEintr = readdir(Verz)) != NULL)
    {
      std::string currentFile = currentDir;
      currentFile += VerzEintr->d_name;
      stat(currentFile.c_str(), &Attribs);

      if (S_ISDIR(Attribs.st_mode) != 0)
      {
        if (strcmp("..", VerzEintr->d_name) != 0 && strcmp(".", VerzEintr->d_name) != 0)
        {
          // Normales Verzeichnis. Einfach darin weitermachen.
          if (removeEmptyDirs(currentFile+"/") == true)
          {
#if DODELETE == 1
            if (rmdir(currentFile.c_str()))
            {
              std::cout << "Error removing directory " << currentFile << "\n";
            }
#else
            std::cout << "Removing directory " << currentFile << "\n";
#endif
          }
          else
            fEmpty = false;
        }
      }
      else
      {
        // Hier ist eine Datei drin, also ist das Verzeichnis nicht leer.
        fEmpty = false;
      }
    }
    closedir(Verz);
  }
  
  return(fEmpty);
}

/**
 * Alle Links betrachten. Wenn ein Link auf eine Datei in 'files' zeigt, 
 * bei diesem Eintrag 'fUsed' auf true setzen.
 */
int SiVeCoRemove::schrittZwei(std::string currentDir)
{
#if DEBUGFIND == 1  
  std::cout << "neueres Verzeichnis: " << currentDir << "\n";
#endif  
  DIR*            Verz;                 // der VerzeichnisStrom
  dirent*         VerzEintr;            // hier werden die Daten aus obigem gegeben
  struct stat     Attribs;              // ein Puffer fr die Dateiattribute

  // relative Verzeichnisangabe erzeugen
  std::string currentDirRel = currentDir.substr(nBaseChars);
#if DEBUGFIND == 1  
  std::cout << "currentDirRel = " << currentDirRel << "\n";
#endif  
  // Wieviele Pfadabschnitte gibt es? In einem neuen String
  // fr jeden ein '../' erzeugen.
  std::string relUp = "";
  int         ups   = 0;
  {
    char* chrPtrEnd = &currentDirRel[currentDirRel.length()-1];
    char* chrPtr    = strchr(&currentDirRel[0], '/');

    while (chrPtr != (char*)0 && chrPtr <= chrPtrEnd)
    {
      relUp += "../";
      ups++;
      chrPtr = strchr(chrPtr, '/')+1;
    }
#if DEBUGFIND == 1  
    std::cout << "up: " << relUp << " (" << ups << ")\n";
#endif
    ups *= 3;
  }
    
  Verz = opendir(currentDir.c_str());
  if (Verz != 0)
  {
    while ((VerzEintr = readdir(Verz)) != NULL)
    {
      std::string currentFile = currentDir;
      currentFile += VerzEintr->d_name;
      lstat(currentFile.c_str(), &Attribs);

      if (S_ISLNK(Attribs.st_mode) != 0)
      {
        // Kommt das Link in der Liste vor?
        std::string relPfad = getLinkTarget(currentFile);
        int         index;

        /* so ist es wenig exakt:
        while (relPfad.find("../") == 0)
          relPfad.replace(0, 3, "");
         */
        // so besser:
        if (relPfad.find(relUp) == 0)
        {
          relPfad = relPfad.substr(ups);
        
          index = findInList(relPfad);
          if (index != -1)
            files[index].fUsed = true;
        }
#if DEBUGFIND == 1
        else
        {
          std::cout << "wird nicht gesucht: " << relPfad << "\n";
        }
#endif        
      }
      else
      {
        if (S_ISDIR(Attribs.st_mode) != 0)
        {
          if (strcmp("..", VerzEintr->d_name) != 0 && strcmp(".", VerzEintr->d_name) != 0)
          {
            // Normales Verzeichnis. Einfach darin weitermachen.
            if (schrittZwei(currentFile+"/") != 0)
            {
              std::cout << "Error\n";
              return(-1);
            }
          }
        }
      }
    }
    closedir(Verz);
  }
  
  return(0);
}

/**
 * Alle wirklichen Dateien des Verzeichnisses in der Liste 'files' merken
 */
int SiVeCoRemove::schrittEins(std::string currentDir)
{
  DIR*            Verz;                 // der VerzeichnisStrom
  dirent*         VerzEintr;            // hier werden die Daten aus obigem gegeben
  struct stat     Attribs;              // ein Puffer fr die Dateiattribute
    
  Verz = opendir(currentDir.c_str());
  if (Verz != 0)
  {
    while ((VerzEintr = readdir(Verz)) != NULL)
    {
      std::string currentFile = currentDir;
      currentFile += VerzEintr->d_name;
      lstat(currentFile.c_str(), &Attribs);

      if (S_ISLNK(Attribs.st_mode) != 0)
      {
#if DODELETE == 1
        if (remove(currentFile.c_str()) != 0)
        {
          std::cout << "Error: unable to remove link " << currentFile << "\n";
          return(-1);
        }
#else        
        std::cout << "Removing " << currentFile << "\n";
#endif        
      }
      else
      {
        if (S_ISDIR(Attribs.st_mode) != 0)
        {
          if (strcmp("..", VerzEintr->d_name) != 0 && strcmp(".", VerzEintr->d_name) != 0)
          {
            // Normales Verzeichnis. Einfach darin weitermachen.
            if (schrittEins(currentFile+"/") != 0)
            {
              std::cout << "Error\n";
              return(-1);
            }
          }
        }
        else
        {
          // Normale Datei. In die Liste aufnehmen.
          addToList(currentFile.substr(nBaseChars));
        }          
      }
    }
    closedir(Verz);
  }
  
  return(0);
}

/**
 * Fgt den string item zur Liste 'files' hinzu
 */
void SiVeCoRemove::addToList(std::string item)
{
  T_Items tmp;
  
  tmp.name  = item;
  tmp.fUsed = false;
  
  if (files.size() == 0)
    files.push_back(tmp);
  else
  {
    unsigned int index = indexInList(item);
    
    if (index == files.size())
      files.push_back(tmp);
    else
      files.insert(files.begin()+index, tmp);
  }
}

/**
 * Zeigt die Liste 'files' an
 */
void SiVeCoRemove::showList()
{
  std::cout << "----------------------\n";
  for (unsigned int cnt=0; cnt<files.size(); cnt++)
  {
    if (files[cnt].fUsed)
      std::cout << " used";
    else
      std::cout << "     ";
    std::cout << "   " << files[cnt].name;
    std::cout << "\n";    
  }
}

/**
 * Liefert -1 zurck, wenn 'item' nicht in der Liste ist,
 * ansonsten dessen Index.
 */
int SiVeCoRemove::findInList(std::string item)
{
  unsigned int index = indexInList(item);
  unsigned int size  = files.size();
  
#if DEBUGFIND == 1  
  std::cout << "suchen: " << item << "\n";
#endif  
  // Wenn das Element vorhanden ist, knnte es an 'index' stehen
  // oder an 'index+1'.
  if (index < size)
    if (item.compare(files[index].name) == 0)
      return(index);
  if (index+1 < size)
    if (item.compare(files[index+1].name) == 0)
      return(index+1);    
  
  // ist nicht vorhanden
  return(-1);
}

/**
 * Liefert den Index zurck, VOR welchem man 'item' einsortieren msste 
 */
int SiVeCoRemove::indexInList(std::string item)
{
#if DEBUGFIND == 1  
  std::cout << "--- Position finden fr " << item << " ------\n";
#endif  
  // Wenn die Liste nur ein Element gro ist, betrachte ich das 
  // als Sonderfall.
  if (files.size() == 1)
  {
    if (strcmp(item.c_str(), files[0].name.c_str()) > 0)
      return(1);
    else
      return(0);
  }
    
  int  step    = files.size() >> 1;
  int  index   = 0;
  int  maxindex= files.size()-1;
  int  nEnde   = 0;
  
  // Ich versuche mich dem Eintrag einigermaen schlau zu nhern.
  // Die Schleife wird beendet, wenn man einen Schritt um eins nach hinten gemacht
  // hat und anschlieend einen Schritt ums eins nach vorne.
  while (nEnde < 2)
  {
#if DEBUGFIND == 1  
    std::cout << index << ", " << step << ", " << nEnde << ": " << files[index].name;
#endif    
    if (strcmp(item.c_str(), files[index].name.c_str()) > 0)
    {
#if DEBUGFIND == 1  
      std::cout << " +\n";
#endif      
      // Ebenfalls beenden, wenn es hinter das letzte Element muss
      if (index == maxindex)
        return(maxindex+1);
      
      index += step;
      
      if (index > maxindex)      
        index = maxindex;      
      
      if (step == 1 && nEnde == 1)
        nEnde = 2;
      else
        nEnde = 0;
    }
    else
    {
#if DEBUGFIND == 1  
      std::cout << " -\n";
#endif      
      // Ebenfalls beenden, wenn es vor das erste Element muss
      if (index == 0)
        return(0);
      
      index -= step;
      if (index < 0)
        index = 0;
      
      if (step == 1)
        nEnde = 1;
    }

    step >>= 1;
    if (step == 0)
    {
      step = 1;
    }
  }
  
#if DEBUGFIND == 1  
  std::cout << "--- Position gefunden ------\n";
#endif  
  return(index);
}
