// -*- mode: c; mode: fold -*-

/*
 * Copyright (C) 2004  Jens Wilhelm Wulf, j dot w dot wulf at gmx dot net
 * Licensed under the GNU General Public License version 2.
 */

#include "tools.h"
#include <fstream>
#include <sys/stat.h>   // Abfragen der Dateiattribute
#include <unistd.h>

#include <sys/wait.h>
#include <stdlib.h>
#include <iostream>
#include <string>
#include <dirent.h>     // fr die Verzeichnisoperationen
#include <sys/types.h>

#include "XMLExtract.h"
#include "filetools.h"

#define DEBUG_TOOLS_CPP 0

bool existDir(std::string DirPfad)
{
  struct stat     Attribs;              // ein Puffer fr die Dateiattribute
  
  if (stat(DirPfad.c_str(), &Attribs))
    return(false);
  
  if (S_ISDIR(Attribs.st_mode) != 0)
    return(true);
  else  
    return(false);
}

bool existDat(std::string DateiPfad) /*{{{*/
{
  std::ifstream testDat;

  if (DateiPfad.length() == 0)
    return(false);
  
  testDat.open((const char*)DateiPfad.c_str());
  
  if (!testDat) 
    return(false);
  else 
  { 
    testDat.close();
    return(true); 
  }
}
/*}}}*/

/**
 * Wohin zeigt ein Link?
 */
std::string getLinkTarget(std::string path)
{
  char*        buf;
  size_t       bufsize = 200;
  unsigned int pos;
  
  buf = (char*)calloc(bufsize, 1);

  while (readlink(path.c_str(), buf, bufsize) == -1)
  {
    bufsize <<= 1;
    std::cout << bufsize << "\n";
    free(buf);
    buf = (char*)calloc(bufsize, 1);
    if (bufsize > 4000)
      return("");
  }
  // Der Inhalt steht jetzt in buf, aber nicht NUL-terminated.
  // Daher muss man calloc benutzen...
  path = buf;
  
  free(buf);
  
  // Evtl. vorkommende "/./" entfernen
  while ((pos = path.find("/./")) != std::string::npos)
  {
    std::string path1 = path.substr(0, pos);
    path  = path.substr(pos+2);
    path  = path1 + path;          
  }
    
  // Evtl. vorkommende "./" zu Beginn entfernen
  if (path.find("./") == 0)
    path  = path.substr(2);
    
  return(path);
}

/**
 * Wandelt einen relativen Pfad in einen absoluten um
 */
std::string relPathToAbs(std::string path)
{
  std::string   path1;
  unsigned int  pos;
  
//  std::cout << "relPathToAbs: " << path << "\n";
  
  while ((pos = path.find("/./")) != std::string::npos)
  {
    path1 = path.substr(0, pos);
    path  = path.substr(pos+2);
    path = path1 + path;          
  }
  
//  std::cout << "relPathToAbs: " << path << "\n";
  
  while ((pos = path.find("../")) != std::string::npos)
  {
    path1 = path.substr(0, pos);
    path  = path.substr(pos+3);

//    std::cout << "path1 " << path1 << "\n";
//    std::cout << "path  " << path << "\n";
    
    // Der vordere Teil des Pfades muss mit '/' beginnen und enden, dazwischen
    // muss mindesten ein Abschnitt sein
    if (path1.length() < 2 || path1[0] != '/' || path1[path1.length()-1] != '/')
    {
//      std::cout << "fehler\n";
      return("");
    }

    path1.replace(path1.length()-1, 1, "");
    while (path1[path1.length()-1] != '/')
      path1.replace(path1.length()-1, 1, "");
    
    path  = path1 + path;    
  }
  
  return(path);
}

/**
 *
 */
std::string getDateString()
{
  char         dt[11];
  time_t       t;
  struct tm*   tt;
  std::string  tmp="";
  
  time(&t);
  tt = localtime(&t);
  strftime(dt, 11, "%Y-%m-%d", tt);
  tmp.append(dt);
  return(tmp);
}

/**
 *
 */
std::string getTimeString()
{
  char         dt[9];
  time_t       t;
  struct tm*   tt;
  std::string  tmp="";
  
  time(&t);
  tt = localtime(&t);
  strftime(dt, 9, "%H:%M.%S", tt);  
  tmp.append(dt);
  return(tmp);
}

/** 
 * Get current version number as int.
 * Returns -1 if there is nothing in the repository.
 */
int getCurrentVersionNumberInt(SimpleXMLTransfer* conf)
{
  DIR*            Verz;                 // der VerzeichnisStrom
  dirent*         VerzEintr;            // hier werden die Daten aus obigem gegeben
  struct stat     Attribs;              // ein Puffer fr die Dateiattribute
  int             maxNum = -1;

  Verz = opendir(conf->attribute("repository").c_str());
  if (Verz != 0) 
  {
    while ((VerzEintr = readdir(Verz)) != NULL)
    {
      std::string tmp = getConfDir(conf, "repository");
      tmp += VerzEintr->d_name;
      stat(tmp.c_str(), &Attribs);
      
      if (S_ISDIR(Attribs.st_mode) != 0 && strcmp("..", VerzEintr->d_name) != 0 && strcmp(".", VerzEintr->d_name) != 0)
      {
        // Convert string to number and compare.
        int tmp = atoi(VerzEintr->d_name);
        if (tmp > maxNum)
          maxNum = tmp;
      }      
    }
    closedir(Verz);
  }
  return(maxNum);
}

std::string getRelPath(std::string refPath,
                       std::string absPath)
{
#if DEBUG_TOOLS_CPP == 1  
  std::cout << "refPath: " << refPath << "\n";
  std::cout << "absPath: " << absPath << "\n";  
#endif  
  if (absPath.find(refPath) == 0)
  {
    // Der einfachste Fall
    
#if DEBUG_TOOLS_CPP == 1  
    std::cout << "./" << absPath.substr(refPath.length()) << "\n";
#endif    
    return("./" + absPath.substr(refPath.length()));
  }
  else
  {
    std::string  basis = refPath;
    std::string  hoch  = "";            
    // Den einen so lange hinten abschneiden, bis er in dem anderen vorkommt.
    while (absPath.find(basis) == std::string::npos && basis.length() > 1)
    {
      basis.replace(basis.length()-1, 1, "");
      while (basis[basis.length()-1] != '/')
        basis.replace(basis.length()-1, 1, "");
        
      hoch.append("../");
    }
    
    // 'Basis' ist jetzt die gemeinsame Basis, 'hoch' enthlt den Weg
    // vom Referenzpfad zur gemeinsamen Basis.
          
#if DEBUG_TOOLS_CPP == 1  
    std::cout << hoch << absPath.substr(basis.length()) << "\n";
#endif    
    return(hoch + absPath.substr(basis.length()));
  }
}

bool canHandleFileVersion(int v)
{
  if (v == 1)
    return(true);
  else
    return(false);
}

int isDerivedFrom(SimpleXMLTransfer* conf,
                  std::string        wdir)
{
  int nNr = -1;
  
  try
  {
    SimpleXMLTransfer* tracking;
    std::string        filename = wdir + conf->attribute("tracking");
    
    tracking = new SimpleXMLTransfer(filename);
    nNr = tracking->attributeAsInt("RepositoryNumber");
    delete tracking;
  }
  catch (XMLException e)
  {
    nNr = -1;
  }
  
  return(nNr);
}

int executeCmd(std::string           cmd,
               std::string           argument)
{
  if (existDat(cmd))
  {    
    std::string path = cmd;
  
    while (path[path.length()-1] != '/')
      path = path.replace(path.length()-1, 1, "");
  
    std::cout << "Executing " << cmd << " " << argument << "\n";
    
    if (chdir(path.c_str()))
    {
      std::cout << "execution of " << cmd << " failed: chdir\n";
      return(-1);
    }

    cmd += " " + argument;
    if (system(cmd.c_str()))
    {
      std::cout << "execution of " << cmd << " failed: system\n";
      return(-1);
    }
    
    return(0);
  }
  else
  {
    std::cout << "Not trying to execute " << cmd << "\n";
    return(-1);
  }
  
}

int setDerivedFrom(SimpleXMLTransfer* conf,
                   std::string        wdir,
                   int                nNr)
{
  if (conf->indexOfAttribute("tracking") >= 0 && conf->attribute("tracking").length() != 0)
  {
    SimpleXMLTransfer*  out = (SimpleXMLTransfer*)0;
    std::ofstream       ofile;  
    std::string         ofilename;
    
    ofilename = wdir + conf->attribute("tracking");
    
    FileTools::makeSureDirExists(FileTools::path(ofilename));
    
    ofile.open(ofilename.c_str());
    if (!ofile)
    {
      std::cout << "Error creating file: " << ofilename << "\n";
      return(-1);
    }
    
    out = new SimpleXMLTransfer();
    out->setName("SiVeCoTracking");
    out->addAttribute("RepositoryNumber", nNr);
    
    // out in ofile schreiben
    out->print(ofile);
    ofile.close();
    delete out;        
  }  
  
  return(0);
}

/**
 * Eine abgesetzte Funktion, damit man leichter mal 
 * eine Formatnderung (bzw. Erweiterung auf mehrere Formate) 
 * einbauen kann.
 */
int createEmptyChangeLog(std::string filename)
{
  std::ofstream       ofile;  
    
  ofile.open(filename.c_str());
  if (!ofile)
  {
    std::cout << "Error creating file: " << filename << "\n";
    return(-1);
  }

  ofile << "<html><body>\n  <ul>\n    <li></li>\n    <li></li>\n    <li></li>\n  </ul>\n</body></html>\n";
  
  ofile.close();
  
  return(0);
}

/**
 * Eine abgesetzte Funktion, damit man leichter mal 
 * eine Formatnderung (bzw. Erweiterung auf mehrere Formate) 
 * einbauen kann.
 * -1: Fehler in Datei
 *  0: Inhalt fehlerfrei ausgelesen
 *  1: Datei nicht vorhanden
 */
int getChangeLogContent(std::string& buf,
                        std::string  filename)
{
  if (existDat(filename))
    return(XMLExtract(buf, filename, "html.body", true));
  else
    return(1);
}

std::string getConfDir(SimpleXMLTransfer* conf,
                       std::string        name)
{
  if (conf->indexOfAttribute(name) < 0)
  {
    std::cerr << "Warning: " << name << " is not defined.\n";
    return("");
  }
  else
  {
    std::string tmp = conf->attribute(name);
    if (tmp.length() == 0)
      return("");
    else
    {
      // Der Pfad muss mit einem '/' enden
      if (tmp[tmp.length()-1] != '/')
        tmp += '/';
      return(tmp);
    }
  }
}
